// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.

#include <tclap/CmdLine.h>

#include <eopcc/cloud_io.hpp>
#include <eopcc/gdal_io.hpp>
#include <eopcc/metrics.hpp>
#include <eopcc/announcer.hpp>
#include <eopcc/registration.hpp>

#include <json/json.h>
#include <fstream>


/// Contains all results from running the selected suite of metrics
class MetricData {
private:
  bool median_error_valid_;
  double median_error_;
  bool rms_error_valid_;
  double rms_error_;

	bool cumul_hist_valid_;
  std::vector<double> cumul_hist_;
  bool error_percentile_valid_;
  std::vector<double> error_precentile_;
public:

	std::vector<double> completeness_sweep;
	std::vector<double> comp_sweep_thresholds;

  double completeness;

  eopcc::Point reg_shift;

  MetricData() {
    median_error_valid_ = false;
    rms_error_valid_ = false;
    cumul_hist_valid_ = false;
    error_percentile_valid_ = false;
  }

  void SetMedianError(double median_error) {
    median_error_valid_ = true;
    median_error_ = median_error;
  }

  void SetRMSError(double rms_error) {
    rms_error_valid_ = true;
    rms_error_ = rms_error;
  }

  void SetCumulativeHistogram(const std::vector<double>& cumul_hist)
  {
	  cumul_hist_valid_ = true;
	  cumul_hist_ = cumul_hist;
  }

  void SetErrorPercentile(const std::vector<double>& error_percentile)
  {
	  error_percentile_valid_ = true;
	  error_precentile_ = error_percentile;
  }

  Json::Value to_json() {
    Json::Value value;
    value["completeness"] = completeness;

    if (rms_error_valid_) {
      value["error_rms"] = rms_error_;
    }

    if (median_error_valid_) {
      value["error_median"] = median_error_;
    }

    if (cumul_hist_valid_) {
	    value["cumulative_histogram"] = Json::Value(Json::arrayValue);
	    for(int it = 0; it < cumul_hist_.size(); it++)
		    value["cumulative_histogram"].append(cumul_hist_[it]);
    }

    if (error_percentile_valid_) {
	    value["error_percentile"] = Json::Value(Json::arrayValue);
	    for(int it = 0; it < error_precentile_.size(); it++)
		    value["error_percentile"].append(error_precentile_[it]);
    }

    if (completeness_sweep.size() > 0) {
	    value["completeness_sweep"] = Json::Value(Json::arrayValue);
	    value["completeness_sweep_thresholds"] = Json::Value(Json::arrayValue);
	    for(int ii = 0; ii < completeness_sweep.size(); ii++) {
		    value["completeness_sweep"].append(completeness_sweep[ii]);
		    value["completeness_sweep_thresholds"].append(comp_sweep_thresholds[ii]);
	    }
    }

	  value["reg_shift"] = Json::Value(Json::arrayValue);
	  value["reg_shift"].append(reg_shift.x);
    value["reg_shift"].append(reg_shift.y);
    value["reg_shift"].append(reg_shift.z);

    return value;
  }
};

/// A job requesting the type of metrics to be executed and the parameters for them
struct MetricsJob {
	eopcc::RegSearchSpace search_space;

	// primary Completeness threshold
  double threshold;

	// Input file and other contestant metadata
  std::string contestant;
  std::string input_file;
  int id;
  std::string md5_sum;

	// Results (after being processed)
  bool valid_results;
  MetricData results;

	// Optional completeness sweep parameters
	bool do_comp_sweep;
	double comp_range_start;
	double comp_range_end;
	double comp_inc;

	MetricsJob() : threshold(0.3), valid_results(false) {
		do_comp_sweep = false;
		comp_range_start = 0.25;
		comp_range_end = 2;
		comp_inc = 0.25;
	}

  void FillFromJson(const Json::Value& json) {
    contestant = json["contestant"].asString();
    input_file = json["submission_file"].asString();
    id = json["submission_id"].asInt();
    md5_sum = json["submission_md5"].asString();
  }

  Json::Value to_json() {
    Json::Value val;
    val["contestant"] = contestant;
    val["input_file"] = input_file;
    val["submission_id"] = id;
    val["submission_md5"] = md5_sum;
    if (valid_results) {
      val["results"] = results.to_json();
    } else {
      val["results"] = "failed to process";
    }

    return val;
  }
};

// Runs a MetricsJob and fills in the MetricData results
void ProcessMetricsJob(MetricsJob& job, const eopcc::GridModel<double>& truth_grid) {
  eopcc::Announcer announ;
  announ.SetNumSteps(4);

  announ.StepStart("Loading " + job.input_file);
  eopcc::PointCloud input;
  eopcc::Load(job.input_file, input);
  announ.StepEnd("Done");

  MetricData metric_data;

  announ.StepStart("Calculating RegisterFromMetrics");
  eopcc::Point shift;

  //shift = eopcc::RegisterFromRmsGrid(job.search_space, input, truth_grid);
  shift = eopcc::RegisterFromMedianGrid(job.search_space, input, truth_grid);

  announ.StepEnd("Done");

  announ.StepStart("Calculating post-registration errors");
  eopcc::GridModel<double> temp_grid(truth_grid.extent(), truth_grid.spacing());
  eopcc::MaxHeightGrid(temp_grid, input, shift);

  metric_data.SetRMSError(eopcc::ZRMSError(temp_grid, truth_grid));
  std::vector<double> buffer(temp_grid.grid.size());
  metric_data.SetMedianError(eopcc::ZMedianError(temp_grid, truth_grid, buffer));
  //eopcc::GridModel<double> error = eopcc::ZRMSErrorGrid(temp_grid, truth_grid);

  //cumulative histogram error
  const double arr[] = {0.5,1,2,5,25};
  int lengthbuf = sizeof(arr) / sizeof(arr[0]);
  std::vector<double> errorTHs(arr, arr + lengthbuf);
  //std::transform(errorTHs.begin(),errorTHs.end(),errorTHs.begin(),
  //               std::bind1st(std::multiplies<double>(),1));//0.000000001));
  metric_data.SetCumulativeHistogram(eopcc::ZCumulativeHistogram(temp_grid, truth_grid, errorTHs));

  //error percentile
  const double arrfraction[] = {0.1,0.25,0.5,0.75,0.9,0.99};
  lengthbuf = sizeof(arrfraction) / sizeof(arrfraction[0]);
  std::vector<double> errorPercentile(arrfraction, arrfraction + lengthbuf);
  metric_data.SetErrorPercentile(eopcc::ZErrorPercentile(temp_grid, truth_grid, errorPercentile));


  announ.StepEnd("Done");

  announ.StepStart("Calculating Completeness with threshold: " + std::to_string(job.threshold));
  metric_data.completeness = eopcc::Completeness(temp_grid, truth_grid, job.threshold);
  announ.StepEnd("Done");

  if (job.do_comp_sweep) {
	  for (double threshold=job.comp_range_start; threshold <= job.comp_range_end; threshold += job.comp_inc) {
		  metric_data.completeness_sweep.push_back(eopcc::Completeness(temp_grid, truth_grid, threshold));
		  metric_data.comp_sweep_thresholds.push_back(threshold);
	  }
  }

  metric_data.reg_shift = shift;


  job.results = metric_data;
  job.valid_results = true;

  //input.translate(shift);
  //eopcc::WriteToLAS(job.input_file + "_registered.las", input);


  return;
}

// Creates a single job from the provided arguments and runs it
MetricData RunMetrics(const std::string& input_file, const std::string& truth_file, double threshold, double do_sweep) {
  eopcc::Announcer announ;
  announ.SetNumSteps(0);

  if (truth_file.empty()) {
    throw std::runtime_error("Missing truth file input argument");
  }

  announ.StepStart("Loading " + truth_file);
  eopcc::GridModel<double> truth_grid = eopcc::LoadFromDEM(truth_file).convertTo<double>();
  announ.StepEnd("Done");

  MetricsJob job;
  job.contestant = "single_run";
  job.id = 0;
  job.md5_sum = "N/A";
  job.input_file = input_file;
  job.threshold = threshold;
  job.do_comp_sweep = do_sweep;

  ProcessMetricsJob(job, truth_grid);

  return job.results;
}

// Runs all the jobs from a batch JSON file
void RunMetricsBatch(const std::string& batch_json, const std::string& args_truth_file) {
    eopcc::Announcer announ;
    announ.SetNumSteps(3);

    std::string truth_file;

    // Open the JSON file
    std::fstream in_file(batch_json, std::ios::in);

    if (!in_file) {
      std::cerr << "Failed to open " << batch_json << std::endl;
      return;
    }

    Json::Value doc;
    in_file >> doc;
    in_file.close();

    // Parse out the top level settings
    double threshold = doc["completeness_threshold"].asDouble();
    double reg_max_shift = doc["reg_max_shift"].asDouble();
    double reg_num_steps = doc["reg_num_steps"].asDouble();

    std::cout << "Using settings: " << threshold << " " << reg_max_shift << " " << reg_num_steps << std::endl;

    if (args_truth_file.empty()) {
      truth_file = doc["truth_file"].asString();
    } else {
      std::cerr << "Warning: overriding truth_file in batch file with command line argument" << std::endl;
      truth_file = args_truth_file;
      doc["truth_file"] = args_truth_file;
    }


    // Load the truth file
    announ.StepStart("Loading " + truth_file);
    eopcc::GridModel<double> truth_grid = eopcc::LoadFromDEM(truth_file).convertTo<double>();
    announ.StepEnd("Done");


    std::vector<MetricsJob> jobs;

    announ.StepStart("Running Batch");
    for (Json::Value& val : doc["entries"]) {
      MetricsJob job;
      job.FillFromJson(val);
      job.threshold = threshold;
      job.search_space = eopcc::RegSearchSpace(reg_max_shift, reg_num_steps);

      try {
        ProcessMetricsJob(job, truth_grid);
      } catch (std::exception& e) {
        std::cerr << "Skipping job, error: " << e.what() << std::endl;
      }
      jobs.push_back(job);
    }
    announ.StepEnd();

    announ.StepStart("Writing Results");
    Json::Value output;
    for (auto job : jobs) {
      output.append(job.to_json());
    }

    std::string results_filename = batch_json.substr(0, batch_json.find_last_of(".")) + "_results.json";

    doc["entries"] = output;

    std::fstream out_file(results_filename, std::ios::out);
    out_file << doc;
    announ.StepEnd();
}


int main(int argc, char* argv[]) {
  try {
    TCLAP::CmdLine cmd("run-metrics command line", ' ', "1.1");

    TCLAP::ValueArg<std::string> batch_input_json("", "batch-input", "json file describing input", false, "", "JSON file");
    TCLAP::ValueArg<std::string> input_filename("i","input", "input point cloud", false,  "", "las/laz file");
    TCLAP::ValueArg<std::string> truth_filename("t","truth", "truth DEM file", false,  "", "tif/tiff file");
    TCLAP::ValueArg<double> compl_threshold("", "cthreshold", "threshold in meters", false, 1.0, "completeness threshold (m)");
    TCLAP::SwitchArg do_sweep("", "completeness-sweep", "enable completeness sweep");

    cmd.add(batch_input_json);
    cmd.add(input_filename);
    cmd.add(truth_filename);
    cmd.add(compl_threshold);
    cmd.add(do_sweep);
    cmd.parse(argc, argv);

    std::string input_file = input_filename.getValue();
    std::string truth_file = truth_filename.getValue();
    std::string batch_input_file = batch_input_json.getValue();
    bool do_sweep_val = do_sweep.getValue();
    if (do_sweep_val) {
	    std::cout << "Completeness sweep is enabled" << std::endl;
    }

    if (input_file != "") {
	    MetricData md = RunMetrics(input_file, truth_file, compl_threshold.getValue(), do_sweep_val);
      std::cout << md.to_json() << std::endl;
      std::string results_filename = input_file + "_results.json";
      Json::Value doc;
      doc["entries"] = md.to_json();

      std::fstream out_file(results_filename, std::ios::out);
      out_file << doc;
    } else if (batch_input_file != "") {
      RunMetricsBatch(batch_input_file, truth_file);
    } else {
      std::cerr << "Either normal input or batch input required" << std::endl;
      return EXIT_FAILURE;
    }


  } catch (std::exception& e) {
    std::cerr << "error: " << e.what() << std::endl;
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}
